from prometheus_client import (
    start_http_server,
    Gauge,
    REGISTRY,
    PROCESS_COLLECTOR,
    PLATFORM_COLLECTOR,
)
from dexus_vault.utils.logger import logger

"""
Define the Prometheus metrics and start the metrics server.
"""

client_create_metric = Gauge("client_create", "Client creation status", ["status"])
client_delete_metric = Gauge("client_delete", "Client deletion status", ["status"])
client_update_metric = Gauge("client_update", "Client update status", ["status"])
client_skip_metric = Gauge("client_skip", "Clients skipped to sync")
incorrect_secrets_metric = Gauge(
    "incorrect_secrets",
    "Number of incorrect secrets specifications in secret engine",
)

# TODO: research if pydantic can cover that
current_state = {
    "incorrect_secrets": 0,
    "clients_skipped": 0,
    "clients_created": 0,
    "clients_deleted": 0,
    "clients_updated": 0,
    "clients_delete_failed": 0,
    "clients_create_failed": 0,
    "clients_update_failed": 0,
}


def reset_state(state):
    """
    Reset the state.
    """
    for key in state.keys():
        state[key] = 0
    logger.debug("State metrics has been reset")


def state_counter(state: dict, response: dict) -> dict:
    """
    Increment the state counter.
    """
    # Count deleted
    if response["operation"] == "delete":
        if response["status"] == "ok":
            state["clients_deleted"] += 1
        elif response["status"] == "failed":
            state["clients_delete_failed"] += 1

    # Count created
    elif response["operation"] == "create":
        if response["status"] == "ok":
            state["clients_created"] += 1
        elif response["status"] == "skipped":
            state["clients_skipped"] += 1
        elif response["status"] == "failed":
            state["clients_create_failed"] += 1

    # Count updated
    elif response["operation"] == "update":
        if response["status"] == "ok":
            state["clients_updated"] += 1
        # that's because 'get' client returned None, but on create we got {alreadyExists: True}
        elif response["status"] == "skipped":
            state["clients_skipped"] += 1
        elif response["status"] == "failed":
            state["clients_create_failed"] += 1

    # Count secrets
    elif response["operation"] == "secret":
        if response["status"] == "ok":
            logger.debug(
                "Secrets are not counted in state yet, check discussions github repo"
            )
        elif response["status"] == "failed":
            state["incorrect_secrets"] += 1
    else:
        logger.debug(f"No state to update on response {response}")

    logger.debug(f"Current updated: {state}")
    return state


def publish_metrics(state: dict) -> None:
    """
    Route the state to correct metric.
    """
    try:
        client_create_metric.labels(status="ok").set(state["clients_created"])
        client_create_metric.labels(status="failed").set(state["clients_create_failed"])
        client_delete_metric.labels(status="ok").set(state["clients_deleted"])
        client_delete_metric.labels(status="failed").set(state["clients_delete_failed"])
        client_update_metric.labels(status="ok").set(state["clients_updated"])
        client_update_metric.labels(status="failed").set(state["clients_update_failed"])
        client_skip_metric.set(state["clients_skipped"])
        incorrect_secrets_metric.set(state["incorrect_secrets"])
        logger.debug(f"Metrics published with state {state}")
    except:
        logger.warning(f"Failed to update state, got value: {state}")
    # we need to reset anyway
    reset_state(state)


def start_metrics_server(config) -> None:
    """
    Start the Prometheus metrics server on the specified port.
    """
    if not config.internal_metrics:
        # Unregister default metrics
        REGISTRY.unregister(PROCESS_COLLECTOR)
        REGISTRY.unregister(PLATFORM_COLLECTOR)
        REGISTRY.unregister(
            REGISTRY._names_to_collectors["python_gc_collections_total"]
        )
        logger.debug("Internal metrics disabled")
    else:
        logger.debug("Internal metrics enabled")

    if config.metrics_enable:
        start_http_server(config.metrics_port)
        logger.info(f"Prometheus metrics server started on port {config.metrics_port}")
        return True

    else:
        logger.info("Prometheus metrics server disabled")
        return False



from typing import Any, Dict, List, Optional
from pydantic import BaseModel, SecretStr, Field, AliasChoices
from pydantic_settings import BaseSettings


class GeneralConfig(BaseSettings):
    log_level: str = "INFO"
    sync_interval: int = 60
    secrets_engine: str = "vault"


class MetricsConfig(BaseSettings):
    metrics_enable: bool = True
    metrics_port: int = 8000
    internal_metrics: bool = False


class DexConfig(BaseSettings):
    client_crt: Optional[str] = None
    client_key: Optional[str] = None
    ca_crt: Optional[str] = None
    dex_grpc_url: str = "127.0.0.1:5557"
    dex_max_retries: int = 20
    dex_retry_wait: int = 3


class VaultConfig(BaseSettings):
    vault_addr: str = "http://127.0.0.1:8200"
    vault_approle_role_id: Optional[str] = None
    vault_approle_secret_id: Optional[str] = None
    vault_approle_secret_path: Optional[str] = None
    vault_kubernetes_role: Optional[str] = None
    vault_kubernetes_jwt_path: str = (
        "/var/run/secrets/kubernetes.io/serviceaccount/token"
    )
    vault_kubernetes_mount_point: str = "kubernetes"
    vault_token: Optional[SecretStr] = None
    vault_cert: Optional[str] = None
    vault_cert_key: Optional[str] = None
    vault_cert_ca: Optional[bool | str] = False
    vault_ldap_username: Optional[str] = None
    vault_ldap_password: Optional[SecretStr] = None
    vault_request_timeout: int = 5
    vault_max_retries: int = 20
    vault_retry_wait: int = 3
    vault_allow_redirect: bool = False
    vault_namespace: Optional[str] = None
    vault_proxies: Optional[Dict[str, Any]] = None
    vault_mount_point: Optional[str] = None
    vault_clients_path: Optional[str] = None
    vault_engine: str = "v2"


class ClientModel(BaseModel):
    # TODO: make function for getting secret from Dex and put to secrets engine
    id: str
    secret: str
    redirect_uris: Optional[List[str]] = Field(
        default=[], validation_alias=AliasChoices("redirect_uris", "redirectUris")
    )
    trusted_peers: Optional[List[str]] = Field(
        default=[], validation_alias=AliasChoices("trusted_peers", "trustedPeers")
    )
    public: bool = False
    name: Optional[str] = ""
    logo_url: Optional[str] = Field(
        default="", validation_alias=AliasChoices("logo_url", "logoUrl")
    )



import time

from pydantic import ValidationError

from dexus_vault.src.dex_processor import DexClient
from dexus_vault.src.vault_processor import VaultClient

from dexus_vault.utils.logger import logger
from dexus_vault.utils.metrics import (
    start_metrics_server,
    current_state,
    state_counter,
    publish_metrics,
)
from dexus_vault.utils.config import (
    GeneralConfig,
    MetricsConfig,
    VaultConfig,
    DexConfig,
    ClientModel,
)


def metrics_server(config):
    """
    Start the Prometheus metrics server.
    """
    try:
        return start_metrics_server(config)
    except ValidationError as error:
        logger.error(f"Failed to validate metrics variables: {error}")
    except Exception as error:
        logger.error(f"Failed to process metrics configuration: {error}")


def sync_dex_clients(dex_client: object, vault_clients: list) -> set:
    """
    Synchronize Dex clients with Vault clients.
    """

    # TODO: make current clients in memory and compare it with vault, to delete clients that are not in vault
    for vault_client in vault_clients:
        try:
            client = ClientModel(**vault_client)

        except ValidationError as error:
            logger.warning(
                f"Secret '{vault_client['id']}' in Vault, missing 'secret' or have incorrect structure"
            )
            logger.debug(f"ValidationError: {error}")
            state_counter(current_state, {"operation": "secret", "status": "failed"})
            continue

        dex_get_client = dex_client.get_dex_client(client_id=client.id)
        if dex_get_client is not None:

            try:
                client_from_dex = ClientModel(**dex_get_client)

            except ValidationError as error:
                logger.warning(
                    f"Client '{client.id}' returned from Dex have incorrect structure {error}"
                )
                continue

            if client.id == client_from_dex.id:
                if client == client_from_dex:
                    logger.debug(f"Client '{client_from_dex.id}' already exist.")
                    state_counter(
                        current_state, {"operation": "update", "status": "skipped"}
                    )
                else:
                    logger.info(
                        f"Detected changes in '{client_from_dex.id}' client configuration, will be recreated"
                    )
                    delete_response = dex_client.delete_dex_client(client_from_dex.id)
                    if delete_response.get("status") == "ok":
                        create_response = dex_client.create_dex_client(client)
                        state_counter(
                            current_state,
                            {
                                "operation": "update",
                                "status": create_response.get("status"),
                            },
                        )
                    else:
                        logger.warning(
                            f"Failed to delete client '{client.id}', response: {delete_response}"
                        )
                        state_counter(
                            current_state, {"operation": "update", "status": "failed"}
                        )
                        continue
        else:
            logger.info(f"Client '{client.id}' not found, will be created")
            response = dex_client.create_dex_client(client)
            state_counter(current_state, response)


def run():
    """
    Main function to run the Dex client and Vault client synchronization.
    """
    general_config = GeneralConfig()
    dex_client = DexClient(config=DexConfig())
    dex_client.dex_waiter()
    metrics_enabled = metrics_server(config=MetricsConfig())

    while True:
        # redefine dex client, to close previous connection, by method __del__
        dex_client = DexClient(config=DexConfig())
        vault_client = VaultClient(config=VaultConfig())
        # get clients from Vault
        client_configs = vault_client.vault_read_secrets()
        sync_dex_clients(dex_client, client_configs)
        logger.info(
            f"Sync completed, next sync after {general_config.sync_interval} seconds"
        )
        # Publish metrics with current state to Prometheus, then reset state
        if metrics_enabled:
            publish_metrics(current_state)
        # Wait for next sync
        time.sleep(general_config.sync_interval)



import time
import grpc

# Import function that transform GRPC Message into Python dict
from google.protobuf.json_format import MessageToDict

# Load GRPC Stub and Methods from pb2 defined using protobuf builder
from dexus_vault.grpc_dexidp.dexidp.api_pb2_grpc import DexStub
import dexus_vault.grpc_dexidp.dexidp.api_pb2 as pb2

from dexus_vault.utils.logger import logger


class DexClient:
    """
    This class defines all logic for operating with Dex gRPC API
    """

    def __init__(self, config):
        # credentials used for Dex connection including GRPC URL
        self.config = config
        # channel credentials obj placeholder
        self.channel = self.dex_grpc_connect()

    # crete grpc connection to Dex
    def dex_grpc_connect(self) -> object:
        """
        Open connection to Dex gRPC
        """

        if self.config.client_crt and self.config.client_key:
            # load credentials from files
            with open(self.config.ca_crt, "r") as ca, open(
                self.config.client_key, "r"
            ) as key, open(self.config.client_crt, "r") as crt:

                # define secure channel with credentials
                self.creds = grpc.ssl_channel_credentials(
                    root_certificates=ca.read(),
                    private_key=key.read(),
                    certificate_chain=crt.read(),
                )
            return grpc.secure_channel(
                target=self.config.dex_grpc_url, credentials=self.creds
            )

        else:
            # insecure channel
            logger.debug(f"Dex client started in insecure channel")
            return grpc.insecure_channel(target=self.config.dex_grpc_url)

    def dex_grpc_close_connection(self):
        """
        Close connection to Dex gRPC
        """
        if self.channel:
            self.channel.close()

    # TODO: research more clean way for retrying connection
    def dex_waiter(self) -> bool:
        """
        Ensure connection to Dex gRPC
        """
        _retry = 0
        while True:
            try:
                logger.info(f"Dex server version {self.get_dex_version()}")
                return True
            except Exception as error:
                if _retry >= self.config.dex_max_retries:
                    raise RuntimeError(f"Could not connect to Dex gRPC server: {error}")
                else:
                    _retry += 1
                    logger.debug(f"Dex gRPC is unavailable, retying...")
                    time.sleep(self.config.dex_retry_wait)

    def get_dex_version(self) -> dict:
        """
        Get Dex version
        """
        dex_request = pb2.VersionReq()
        response = DexStub(self.channel).GetVersion(dex_request)
        return MessageToDict(response)

    def get_dex_client(self, client_id: str) -> dict | None:
        """
        Get Dex Client by id
        """
        dex_request = pb2.GetClientReq()
        dex_request.id = client_id

        try:
            response = DexStub(self.channel).GetClient(dex_request)
            return MessageToDict(response).get("client")

        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.UNKNOWN:
                logger.debug(
                    f"Dex gRPC response: {rpc_error.details()}, client {client_id}"
                )
            else:
                logger.warning(
                    f"Dex gRPC error code: {rpc_error.code()} with details {rpc_error.details()}"
                )

    def create_dex_client(self, client: dict) -> dict | None:
        """
        Create OIDC client in Dex
        """
        try:
            # define MessageReq parameter for gRPC
            request = pb2.CreateClientReq()
            request.client.id = client.id
            request.client.secret = client.secret
            request.client.redirect_uris.extend(client.redirect_uris)
            request.client.trusted_peers.extend(client.trusted_peers)
            request.client.public = client.public
            request.client.name = client.name
            request.client.logo_url = client.logo_url
            # send request to Dex gRPC and store response
            response = MessageToDict(DexStub(self.channel).CreateClient(request))
            # check if client was created
            if response.get("client", None) is not None:
                client_id = response.get("client").get("id")
                logger.info(f"Created new Dex client '{client_id}'")
                return {
                    "client": client_id,
                    "operation": "create",
                    "status": "ok",
                    "response": response,
                }
            # check if client already exists returned from Dex
            elif response.get("alreadyExists", None) is not None:
                logger.info(
                    f"Client {client.id} already exists, check Vault configs for duplicates"
                )
                return {
                    "client": client.id,
                    "operation": "create",
                    "status": "skipped",
                    "response": response,
                }
            # check if there was an error in response
            else:
                logger.warning(f"Dex gRPC response: {response}")
                return {
                    "client": client.id,
                    "operation": "delete",
                    "status": "failed",
                    "response": response,
                }
        # catch any exception and log it
        except Exception as error:
            logger.error(f"Failed to create client {client.id}")
            logger.error(f"Dex gRPC response: {error}")
            return {
                "client": client.id,
                "operation": "delete",
                "status": "failed",
                "response": error,
            }

    def delete_dex_client(self, client_id: str) -> None:
        """
        Delete OIDC client in Dex
        """
        try:
            dex_request = pb2.DeleteClientReq()
            dex_request.id = client_id

            # Returns None(means deleted) or {'notFound': True}, in other case it will raise an exception
            response = MessageToDict(DexStub(self.channel).DeleteClient(dex_request))
            if response.get("notFound", None) is not None:
                logger.warning(f"Client '{client_id}' not found")
                return {
                    "client": client_id,
                    "operation": "delete",
                    "status": "skipped",
                    "response": response,
                }

            else:
                logger.info(f"client {client_id} was deleted")
                return {
                    "client": client_id,
                    "operation": "delete",
                    "status": "ok",
                    "response": response,
                }

        except Exception as error:
            logger.error(f"Failed to delete client {client_id}")
            logger.error(f"Dex gRPC response: {error}")
            return {
                "client": client_id,
                "operation": "delete",
                "status": "failed",
                "response": error,
            }

    def __del__(self):
        """
        Every time DexClient object deletes, this will close gRPC connection
        """
        self.dex_grpc_close_connection()
        logger.debug(f"Dex connection closed")



import sys
import time
import requests

from hvac.api.auth_methods import Kubernetes
import hvac

from dexus_vault.utils.logger import logger


class VaultClient:
    """
    This Class represents methods used inside dexus_vault, nothing new, just wrap hvac
    """

    def __init__(self, config: dict):
        self.config = config
        self.client = self.login_to_client()

    def _check_vault_status(self, client: object) -> None:
        """
        Function validates if Vault is up and initialized
        """
        # TODO: research more clean way to check if Vault is initialized
        for _ in range(self.config.vault_max_retries):
            try:
                response = client.sys.read_health_status(method="GET")
                # check response type, in some cases it is not requests.Response
                if isinstance(response, requests.Response):
                    status = response.json()
                else:
                    status = response
                if status["initialized"]:
                    logger.debug(f"Vault {self.config.vault_addr} is initialized")
                    return True
                else:
                    logger.warning(
                        f"Vault {self.config.vault_addr} is not initialized {status}"
                    )
            except requests.exceptions.ConnectionError:
                logger.warning(f"Vault {self.config.vault_addr} connection failed")
            time.sleep(self.config.vault_addr)
        logger.error(f"Vault {self.config.vault_addr} unreachable, exiting...")
        sys.exit(1)

    def _check_if_vault_auth(self, client: object, auth_method: str) -> None:
        """
        Function validates if we successfully authenticated to Vault
        """
        if client.is_authenticated():
            logger.debug(
                f"Authenticated to Vault {self.config.vault_addr} via {auth_method} auth method"
            )
        else:
            raise RuntimeError(
                f"Failed to authenticate to Vault {self.config.vault_addr} via {auth_method} auth method "
            )

    def login_to_client(self) -> object:
        """
        Define auth method for Vault and authenticate
        """
        client = hvac.Client(url=self.config.vault_addr)
        self._check_vault_status(client)

        if self.config.vault_approle_role_id:
            auth_method = "approle"

            if self.config.vault_approle_secret_path is not None:
                with open(self.config.vault_approle_secret_path, "r") as secret_id:
                    client.auth.approle.login(
                        role_id=self.config.vault_approle_role_id,
                        secret_id=secret_id.read(),
                    )
            else:
                client.auth.approle.login(
                    role_id=self.config.vault_approle_role_id,
                    secret_id=self.config.vault_approle_secret_id,
                )

        elif self.config.vault_kubernetes_role is not None:
            # POST /auth/{mount_point}/login
            auth_method = "kubernetes"

            with open(self.config.vault_kubernetes_jwt_path, "r") as jwt:
                # it is required to have params order mp, role, jwt
                Kubernetes(client.adapter).login(
                    mount_point=self.config.vault_kubernetes_mount_point,
                    role=self.config.vault_kubernetes_role,
                    jwt=jwt.read(),
                )

        elif self.config.vault_ldap_username and self.config.vault_ldap_password:
            auth_method = "ldap"
            client.auth.ldap.login(
                username=self.config.vault_ldap_username,
                password=self.config.vault_ldap_password.get_secret_value(),
            )

        elif self.config.vault_cert and self.config.vault_cert_key:
            auth_method = "cert"
            if self.config.vault_cert_ca:
                with open(self.config.vault_cert_ca, "r") as ca:
                    client = hvac.Client(
                        url=self.config.vault_addr,
                        cert=(self.config.vault_cert, self.config.vault_cert_key),
                        verify=ca,
                    )
            else:
                client = hvac.Client(
                    url=self.config.vault_addr,
                    token=self.config.vault_token.get_secret_value(),
                    cert=(self.config.vault_cert, self.config.vault_cert_key),
                    verify=(self.config.vault_cert_ca),
                )
                logger.debug(
                    f"Vault client auth method: {auth_method}, but CA verification is disabled"
                )

        elif self.config.vault_token is not None:
            auth_method = "token"
            client.token = self.config.vault_token.get_secret_value()

        else:
            raise KeyError(f"Vault auth method is not specified, or not supported")

        self._check_if_vault_auth(client, auth_method)
        return client

    def vault_list_secrets(self) -> list:
        """
        List secrets from Vault by path
        """
        try:
            if self.config.vault_engine == "v1":
                response = self.client.secrets.kv.v1.list_secrets(
                    self.config.vault_clients_path,
                    mount_point=self.config.vault_mount_point,
                )

            elif self.config.vault_engine == "v2":
                response = self.client.secrets.kv.v2.list_secrets(
                    self.config.vault_clients_path,
                    mount_point=self.config.vault_mount_point,
                )
            return response["data"]["keys"]

        except hvac.exceptions.Forbidden as error:
            logger.error(
                f"Permissions denied for listing secrets by path {self.config.vault_clients_path}, hvac details: {error}"
            )
            logger.info(
                f"For Vault kv-v2 engine, make sure you have 'list' permission on 'metadata' path"
            )
            return []

        except Exception as error:
            logger.error(
                f"Error listing secrets from Vault, via path {self.config.vault_clients_path}, more details: {error}"
            )
            return []

    def vault_read_secret(self, secret_path: str):
        """
        Read specified secret from Vault
        """
        try:
            if self.config.vault_engine == "v1":
                response = self.client.secrets.kv.v1.read_secret(
                    path=f"{self.config.vault_clients_path}/{secret_path}",
                    mount_point=self.config.vault_mount_point,
                )
                return response["data"]

            elif self.config.vault_engine == "v2":
                response = self.client.secrets.kv.read_secret_version(
                    path=f"{self.config.vault_clients_path}/{secret_path}",
                    mount_point=self.config.vault_mount_point,
                )
                return response["data"]["data"]

        except hvac.exceptions.Forbidden as error:
            logger.error(
                f"Permissions denied for reading secret by path {self.config.vault_clients_path}/{secret_path}, hvac details: {error}"
            )
            return None

        except Exception as error:
            logger.error(f"Error reading secret {secret_path} from Vault: {error}")
            return None

    def vault_read_secrets(self) -> list:
        """
        Combine list_secrets with read_secret
        """
        _client_config = []

        for secret in self.vault_list_secrets():
            config = self.vault_read_secret(secret_path=secret)

            if config is not None and config != {}:
                # set Vault secret name as Dex client id if "id" is not set in Vault secret
                if config.get("id", None) is None:
                    config["id"] = secret
                    logger.debug(f"Client id is set to secret name '{secret}'")

                _client_config.append(config)
            # to avoid empty secrets
            else:
                logger.warning(
                    f"Empty secret '{secret}' or insufficient permissions in Vault to read it, skipping..."
                )

        return _client_config



import sys
import logging

from dexus_vault.utils.config import GeneralConfig

general_config = GeneralConfig()

# Define a format for your logs
log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

# Create a stream handler
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setLevel(general_config.log_level)
stream_handler.setFormatter(logging.Formatter(log_format))

# Get a logger
logger = logging.getLogger("dexus_vault")
logger.setLevel(general_config.log_level)

# Add the handlers to the logger
logger.addHandler(stream_handler)
# logger.addHandler(file_handler)

# Set log level for external libraries
logging.getLogger("httpx").setLevel(logging.WARNING)
